/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.utils;

import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.fortiss.tooling.base.layout.DefaultLayoutConstants;
import org.fortiss.tooling.base.layout.LayoutKeyConstants;
import org.fortiss.tooling.base.model.layout.Angle;
import org.fortiss.tooling.base.model.layout.Dimension;
import org.fortiss.tooling.base.model.layout.EOrientation;
import org.fortiss.tooling.base.model.layout.ILayoutData;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.Offset;
import org.fortiss.tooling.base.model.layout.OffsetOrientation;
import org.fortiss.tooling.base.model.layout.Point;
import org.fortiss.tooling.base.model.layout.Points;
import org.fortiss.tooling.base.utils.AngleUtils;
import org.fortiss.tooling.base.utils.DimensionUtils;
import org.fortiss.tooling.base.utils.LayoutModelElementFactory;
import org.fortiss.tooling.base.utils.OffsetOrientationUtils;
import org.fortiss.tooling.base.utils.PointUtils;
import org.fortiss.tooling.base.utils.PointsUtils;

/**
 * Utility methods for {@link ILayoutedModelElement} and {@link ILayoutData}.
 * This class provides access to layout data using the default keys defined in
 * {@link LayoutKeyConstants}. Changing some layout data is done by replacing
 * the layout data object under a given key, which again results in a EMF
 * notification on the {@link ILayoutedModelElement}. Therefore a single change
 * notification listener on the layouted element can be used.
 * 
 * @author hoelzl
 * @author wandinger
 * @author hummel
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: DC17063D1F5AA00AD9D32A1CB6637542
 */
public class LayoutDataUtils {

	/** Returns the bounds rectangle of the given layouted element. */
	public static Rectangle getNodeBounds(ILayoutedModelElement layouted) {
		Rectangle result = new Rectangle();
		Point p = getNodePosition(layouted);
		result.setLocation(p.getX(), p.getY());
		Dimension d = getNodeSize(layouted);
		result.setSize(d.getWidth(), d.getHeight());
		return result;
	}

	/** Sets the bounds rectangle of the given layouted node. */
	public static void setNodeBounds(ILayoutedModelElement layouted,
			Rectangle bounds) {
		setNodePosition(layouted, bounds.x, bounds.y);
		setNodeSize(layouted, bounds.width, bounds.height);
	}

	/** Returns the position of the given layouted node. */
	public static Point getNodePosition(ILayoutedModelElement layouted) {
		return PointUtils.getPosition(layouted,
				LayoutKeyConstants.NODE_POSITION);
	}

	/** Sets the layouted node's position. */
	public static void setNodePosition(ILayoutedModelElement layouted, int x,
			int y) {
		PointUtils.setPoint(layouted, LayoutKeyConstants.NODE_POSITION, x, y);
	}

	/**
	 * Returns the size of the given layouted node. If no dimension is stored in
	 * the given element a new dimension with the default size is returned.
	 */
	public static Dimension getNodeSize(ILayoutedModelElement layouted) {
		Dimension size = DimensionUtils.getDimension(layouted,
				LayoutKeyConstants.NODE_DIMENSION);

		if (size == null) {
			Dimension result = LayoutModelElementFactory.createDimension(
					DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_WIDTH,
					DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_HEIGHT,
					LayoutKeyConstants.NODE_DIMENSION);
			return result;
		}
		return EcoreUtil.copy(size);
	}

	/** Sets the layouted node's size. */
	public static void setNodeSize(ILayoutedModelElement layouted, int w, int h) {
		DimensionUtils.setDimension(layouted,
				LayoutKeyConstants.NODE_DIMENSION, w, h);
	}

	/** Moves the given node. */
	public static void moveNode(ILayoutedModelElement layouted, int dx, int dy) {
		Point p = LayoutDataUtils.getNodePosition(layouted);
		LayoutDataUtils.setNodePosition(layouted, p.getX() + dx, p.getY() + dy);
	}

	/** Moves and resizes the given node. */
	public static void moveAndResizeNode(ILayoutedModelElement layouted,
			int dx, int dy, int dw, int dh, int minWidth, int minHeight) {
		Point pos = LayoutDataUtils.getNodePosition(layouted);
		int posX = pos.getX() + dx;
		int posY = pos.getY() + dy;

		Dimension dim = LayoutDataUtils.getNodeSize(layouted);
		int dimW = dim.getWidth() + dw;
		int dimH = dim.getHeight() + dh;

		if (dimW < minWidth) {
			if (dx > 0) {
				posX = posX - minWidth - dimW;
			}
			dimW = minWidth;
		}

		if (dimH < minHeight) {
			if (dy > 0) {
				posY = posY - minHeight - dimH;
			}
			dimH = minHeight;
		}

		LayoutDataUtils.setNodePosition(layouted, posX, posY);
		LayoutDataUtils.setNodeSize(layouted, dimW, dimH);
	}

	/** Returns the layouted connector's position. */
	public static Point getConnectorPosition(ILayoutedModelElement layouted) {
		return PointUtils.getPosition(layouted,
				LayoutKeyConstants.CONNECTOR_POSITION);
	}

	/** Sets the layouted connector's position. */
	public static void setConnectorPosition(ILayoutedModelElement layouted,
			int x, int y) {
		PointUtils.setPoint(layouted, LayoutKeyConstants.CONNECTOR_POSITION, x,
				y);
	}

	/** Returns the layouted connector's orientation. */
	public static OffsetOrientation getConnectorOffsetOrientation(
			ILayoutedModelElement layouted) {
		OffsetOrientation oo = OffsetOrientationUtils.getOffsetOrientation(
				layouted, LayoutKeyConstants.CONNECTOR_OFFSET_ORIENTATION);
		if (oo == null) {
			return LayoutModelElementFactory.createOffsetOrientation(0,
					EOrientation.NORTH,
					LayoutKeyConstants.CONNECTOR_OFFSET_ORIENTATION);
		}
		return oo;
	}

	/** Sets the layouted connector's offset and orientation. */
	public static void setConnectorOffsetOrientation(
			ILayoutedModelElement layouted, OffsetOrientation orientation) {
		OffsetOrientationUtils.setOffsetOrientation(layouted,
				LayoutKeyConstants.CONNECTOR_OFFSET_ORIENTATION,
				orientation.getOffset(), orientation.getOrientation());
	}

	/** Returns the layouted connector's offset. */
	public static Offset getConnectorOffset(ILayoutedModelElement layouted) {
		OffsetOrientation offset = OffsetOrientationUtils.getOffsetOrientation(
				layouted, LayoutKeyConstants.CONNECTOR_OFFSET);
		if (offset == null) {
			return LayoutModelElementFactory.createOffsetOrientation(0,
					EOrientation.NORTH, LayoutKeyConstants.CONNECTOR_OFFSET);
		}
		return offset;
	}

	/** Sets the layouted connector's angle. */
	public static void setConnectorAngle(ILayoutedModelElement layouted,
			double angle) {
		AngleUtils
				.setAngle(layouted, LayoutKeyConstants.CONNECTOR_ANGLE, angle);
	}

	/** Returns the layouted connector's angle. */
	public static Angle getConnectorAngle(ILayoutedModelElement layouted) {
		Angle angle = AngleUtils.getAngle(layouted,
				LayoutKeyConstants.CONNECTOR_ANGLE);
		if (angle == null) {
			return LayoutModelElementFactory.createAngle(0.0,
					LayoutKeyConstants.CONNECTOR_ANGLE);
		}
		return angle;
	}

	/** Returns the angle as double value (0 if not existing). */
	public static double getConnectorAngleAsDouble(ILayoutedModelElement lobject) {
		return getConnectorAngle(lobject).getAngle();
	}

	/** Return the absolute position for a given offset, orientation and insets. */
	public static Point getAbsolute(OffsetOrientation oo, Dimension parentSize,
			int parentInsets) {
		Point pos = LayoutModelElementFactory.createPoint(0, 0, null);
		int offsetValue = oo.getOffset();
		switch (oo.getOrientation()) {
		case NORTH:
			pos.setX(offsetValue);
			pos.setY(0);
			break;
		case WEST:
			pos.setX(0);
			pos.setY(offsetValue);
			break;
		case SOUTH:
			pos.setX(offsetValue);
			pos.setY(parentSize.getHeight() - 2 * parentInsets);
			break;
		case EAST:
			pos.setX(parentSize.getWidth() - 2 * parentInsets);
			pos.setY(offsetValue);
			break;
		}
		return pos;
	}

	/** Returns the {@link Points} layout data object for the given connection. */
	public static Points getConnectionPoints(ILayoutedModelElement lobject) {
		return PointsUtils.getPoints(lobject,
				LayoutKeyConstants.CONNECTION_POINTS);
	}

	/** Adds the given x/y point to the layout data of the given element. */
	public static void addConnectionPoint(ILayoutedModelElement lobject,
			int index, int x, int y) {
		Points existingPoints = getConnectionPoints(lobject);
		Point newPoint = LayoutModelElementFactory.createPoint(x, y, null);
		existingPoints.getPointsList().add(index, newPoint);
		PointsUtils.setPoints(lobject, LayoutKeyConstants.CONNECTION_POINTS,
				existingPoints);
	}

	/** Changes the point of the layout data at the given index. */
	public static void changeConnectionPoint(ILayoutedModelElement lobject,
			int index, int x, int y) {
		Points pts = getConnectionPoints(lobject);
		Point np = LayoutModelElementFactory.createPoint(x, y, null);
		pts.getPointsList().set(index, np);
		PointsUtils.setPoints(lobject, LayoutKeyConstants.CONNECTION_POINTS,
				pts);
	}

	/** Changes all the point layout data of the given element. */
	public static void changeAllConnectionPoints(ILayoutedModelElement lobject,
			int dx, int dy) {
		Points pts = getConnectionPoints(lobject);
		Points newPoints = LayoutModelElementFactory.createPoints(null);
		for (Point p : pts.getPointsList()) {
			Point np = LayoutModelElementFactory.createPoint(p.getX() + dx,
					p.getY() + dy, null);
			newPoints.getPointsList().add(np);
		}
		PointsUtils.setPoints(lobject, LayoutKeyConstants.CONNECTION_POINTS,
				newPoints);
	}

	/**
	 * Removes the point at the given index from the layout data of the given
	 * element.
	 */
	public static void removeConnectionPoint(ILayoutedModelElement lobject,
			int index) {
		Points pts = getConnectionPoints(lobject);
		pts.getPointsList().remove(index);
		PointsUtils.setPoints(lobject, LayoutKeyConstants.CONNECTION_POINTS,
				pts);

	}
}
